#!/usr/bin/ruby

=begin
 | ---[ PROGRAM DESCRIPTION ]---------------------------------------------------
 | This program runs as a daemon and periodically checks a scheduling table to 
 | find jobs that must be run.  The syntax used in the scheduling table is 
 | related to the CRON formatting.  The scheduler is designed to be fault
 | tolerant and respectful of resources.
 |
 | ---[ INFORMATION ]-----------------------------------------------------------
 |     Author: Ira Burton [ ira.burton@gmail.com ]
 |  Copyright: Copyright (c) 2009. Ira Burton
 |    License: Please check with authors
=end

# Log file in rails log folder (collector.log)
require 'rubygems'
require 'daemons'
require 'active_record' 
require 'yaml'
require 'log4r'

# Cron calculation constants used for cronExpand
CronHR  = 1
CronMIN = 2
CronMON = 4
CronDOW = 8
CronDOM = 16

# Return an array containing all the valid values for a cron field expression
# 
def cronExpand(value, type)
  expansion = []
  if value.strip != '*' then
    value.split(',').each do |v1|
      v1 = v1.strip.split('-')
      if v1.size == 1 then
        expansion.push v1[0].strip.to_i
      else
        v1[0] = v1[0].strip.to_i
        v1[1] = v1[1].strip.to_i
        v1.sort!
        v1[0].upto v1[1] do |i| expansion.push(i) end          
      end
    end    
  else
    0.upto 23 do |i| expansion.push(i) end if type == CronHR
    0.upto 59 do |i| expansion.push(i) end if type == CronMIN
    1.upto 12 do |i| expansion.push(i) end if type == CronMON
    0.upto 6  do |i| expansion.push(i) end if type == CronDOW
    1.upto 31 do |i| expansion.push(i) end if type == CronDOM
  end
  return expansion
end

# Get the next run time based on a cron schedule 
def cronNextTime(baseTime, mon, dom, hr, min, dow)
  # Make sure the base time has no seconds or fractions of seconds
  baseTime = Time.local(baseTime.year, baseTime.month, baseTime.day, baseTime.hour, baseTime.min, 0)
  
  # Get the expansions of the cron entries
  hrExp = cronExpand(hr, CronHR)
  minExp = cronExpand(min, CronMIN)
  monExp = cronExpand(mon, CronMON)
  dowExp = cronExpand(dow, CronDOW)
  domExp = cronExpand(dom, CronDOM)
  
  baseTime += 60
  while !hrExp.member?(baseTime.hour) or 
        !minExp.member?(baseTime.min) or 
        !monExp.member?(baseTime.mon) or
        !domExp.member?(baseTime.day) or
        !dowExp.member?(baseTime.wday) do          
    baseTime += 60
  end
  
  return baseTime
  
end

# Find out where collector.rb is located
BASEDIR = File.dirname(__FILE__) + '/../'
FULLDIR = File.dirname(File.expand_path(__FILE__)) + '/../' 
 
# Grab the scheduler class
# require BASEDIR + 'app/models/pull_schedule'

# Grab the parameter class
# require BASEDIR + 'app/models/parameter'

# Grab the collector classes (files that start with col_)
Dir.new(BASEDIR + 'app/models/').entries.each do |file|
  if file.downcase.split('.')[-1] == 'rb' then
    require BASEDIR + 'app/models/' + file
  end
end

# Connect to the database
dbconfig = YAML.load_file("config/database.yml")['production']
ActiveRecord::Base.establish_connection( dbconfig )

# Now split off the daemon portion of the app
Daemons.run_proc('collector', {:dir_mode => :normal, :dir => BASEDIR + 'log/'}) do
 
  # Set up the logger inside the Daemon 
  mylog = Log4r::Logger.new 'mylog'
  mylog.level = Log4r::INFO
  logfile = Log4r::FileOutputter.new 'logfile', :filename => "#{FULLDIR}log/collector.log"
  mylog.add(logfile)
  logfile.formatter = Log4r::PatternFormatter.new(:pattern => "[%l] %d :: %m")
  Log4r::Logger['mylog'].info "Starting collection daemon"
  
  # Loop forever
  loop do   
    begin
      # Grab any schedule items that should have run already
      Log4r::Logger['mylog'].debug 'Loading Schedule'
      schedItems = PullSchedule.find(:all, :conditions => 'nextrun < NOW()')
    rescue => e
      if e.to_s =~ /away/
        Log4r::Logger['mylog'].error "     ... Reconnecting to database"
        ActiveRecord::Base.establish_connection( dbconfig ) and retry
      else
        schedItems = []
        Log4r::Logger['mylog'].error "     ... Error loading schedule"
      end
    end      

    schedItems.each do |item|        
      begin          
        # Record the time the collection started.
        item.lastrun =  Time.now
        
        if item.collector.active then
          # Execute the collector's gather method
          Log4r::Logger['mylog'].info "Activating #{item.collector.name} collector"
          Kernel.const_get(item.collector.name).gather(item.parameters)
          Log4r::Logger['mylog'].info "     ... updating schedule"
        else
          Log4r::Logger['mylog'].info "Skipping #{item.collector.name} collector, it is not active"
        end
      rescue
        Log4r::Logger['mylog'].error "     ... gather failed."
        puts $!
      end
      begin
        # Look for the next scheduled time that is after right now
        while item.nextrun < Time.now() do 
          item.nextrun = cronNextTime(item.nextrun, item.cron_month, item.cron_dom, item.cron_hour, item.cron_min, item.cron_dow)
        end
      
        item.save
        Log4r::Logger['mylog'].info "     ... complete"
      rescue
        Log4r::Logger['mylog'].error "     ... schedule update failed"
      end      
    end
    sleep(5)
  end
end